/*
	SDL - Simple DirectMedia Layer
	Copyright (C) 1997-2006 Sam Lantinga

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

	Tantric, 2009; Extrems, 2012
*/
#include "SDL_config.h"

#include "../SDL_sysvideo.h"
#include "../SDL_pixels_c.h"
#include <malloc.h>
#include <gccore.h>
#include <ogc/usbmouse.h>
#include <wiikeyboard/keyboard.h>
#include <wiiuse/wpad.h>

#include "SDL_wiivideo.h"
#include "SDL_wiievents_c.h"

static int WII_VideoInit(_THIS, SDL_PixelFormat *vformat);
static SDL_Rect **WII_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags);
static SDL_Surface *WII_SetVideoMode(_THIS, SDL_Surface *current, int width, int height, int bpp, Uint32 flags);
static void WII_UpdateRects(_THIS, int numrects, SDL_Rect *rects);
static int WII_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors);
static void WII_VideoQuit(_THIS);

GXRModeObj *vmode;
static void *xfb;

static GXTexObj texobj[2];
static GXTlutObj tlutobj[2];

static Uint8 gp_fifo[GX_FIFO_MINSIZE] ATTRIBUTE_ALIGN(32);
static Uint16 tlutmem[2][256] ATTRIBUTE_ALIGN(32);
static Sint16 square[16] ATTRIBUTE_ALIGN(32);

static void SetupGX(void)
{
	Mtx44 projection;
	
	GX_Init(gp_fifo, GX_FIFO_MINSIZE);
	
	GX_SetViewport(0.0, 0.0, vmode->fbWidth, vmode->xfbHeight, 0.0f, 1.0f);
	GX_SetDispCopyDst(vmode->fbWidth, vmode->xfbHeight);
	GX_SetCopyFilter(GX_FALSE, NULL, GX_TRUE, vmode->vfilter);
	GX_SetZMode(GX_TRUE, GX_ALWAYS, GX_TRUE);
	
	GX_ClearVtxDesc();
	
	GX_SetVtxDesc(GX_VA_POS, GX_INDEX8);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_POS, GX_POS_XY, GX_S16, 0);
	GX_SetArray(GX_VA_POS, square, sizeof(s16) * 2);
	
	GX_SetVtxDesc(GX_VA_TEX0, GX_INDEX8);
	GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX0, GX_TEX_ST, GX_S16, 0);
	GX_SetArray(GX_VA_TEX0, square, sizeof(s16) * 2);
	
	GX_SetTexCoordScaleManually(GX_TEXCOORD0, GX_ENABLE, 1, 1);
	
	guOrtho(projection, -vmode->xfbHeight, vmode->xfbHeight,
		-vmode->fbWidth, vmode->fbWidth, 0.0f, 1.0f);
	GX_LoadProjectionMtx(projection, GX_ORTHOGRAPHIC);
}

static void DrawInit(_THIS, SDL_Surface *surface)
{
	square[0] = square[6] = -surface->w;
	square[2] = square[4] =  surface->w;
	square[1] = square[3] = -surface->h;
	square[5] = square[7] =  surface->h;
	
	square[10] = square[12] = surface->w;
	square[13] = square[15] = surface->h;
	
	DCFlushRange(square, sizeof(square));
	GX_InvVtxCache();
	
	GX_SetNumTevStages(2);
	GX_SetTevColor(GX_TEVREG0, (GXColor){0x00, 0x00, 0xFF, 0xFF});
	GX_SetTevSwapModeTable(GX_TEV_SWAP1, GX_CH_GREEN, GX_CH_ALPHA, GX_CH_ALPHA, GX_CH_ALPHA);
	
	GX_SetTevOrder(GX_TEVSTAGE0, GX_TEXCOORD0, GX_TEXMAP0, GX_COLORNULL);
	GX_SetTevColorIn(GX_TEVSTAGE0, GX_CC_TEXC, GX_CC_ZERO, GX_CC_C0, GX_CC_ZERO);
	GX_SetTevSwapMode(GX_TEVSTAGE0, GX_TEV_SWAP0, GX_TEV_SWAP1);
	
	GX_SetTevOrder(GX_TEVSTAGE1, GX_TEXCOORD0, GX_TEXMAP1, GX_COLORNULL);
	GX_SetTevColorIn(GX_TEVSTAGE1, GX_CC_ZERO, GX_CC_TEXC, GX_CC_C0, GX_CC_CPREV);
	GX_SetTevAlphaIn(GX_TEVSTAGE1, GX_CA_ZERO, GX_CA_TEXA, GX_CA_A0, GX_CA_ZERO);
	GX_SetTevSwapMode(GX_TEVSTAGE1, GX_TEV_SWAP0, GX_TEV_SWAP2);
	
	GX_InitTlutObj(&tlutobj[0], tlutmem[0], GX_TL_IA8, 256);
	GX_InitTexObjCI(&texobj[0], this->hidden->buffer, this->hidden->w, this->hidden->h,
		GX_TF_CI8, GX_CLAMP, GX_CLAMP, GX_FALSE, GX_TLUT0);
	
	GX_InitTlutObj(&tlutobj[1], tlutmem[1], GX_TL_IA8, 256);
	GX_InitTexObjCI(&texobj[1], this->hidden->buffer, this->hidden->w, this->hidden->h,
		GX_TF_CI8, GX_CLAMP, GX_CLAMP, GX_FALSE, GX_TLUT1);
}

static void VSyncCB(Uint32 retrace)
{
	int dxs, dys;
	for (dys = 0; dys < vmode->xfbHeight; dys += vmode->efbHeight) {
		for (dxs = 0; dxs < vmode->fbWidth; dxs += 640) {
			GX_SetDispCopySrc(0, 0, MIN(vmode->fbWidth - dxs, 640), vmode->efbHeight);
			GX_SetScissor(dxs, dys, MIN(vmode->fbWidth - dxs, 640), vmode->efbHeight);
			GX_SetScissorBoxOffset(dxs, dys);
			
			GX_Begin(GX_QUADS, GX_VTXFMT0, 4);
				GX_Position1x8(0); GX_TexCoord1x8(4);
				GX_Position1x8(1); GX_TexCoord1x8(5);
				GX_Position1x8(2); GX_TexCoord1x8(6);
				GX_Position1x8(3); GX_TexCoord1x8(7);
			GX_End();
			
			GX_CopyDisp(xfb + (dys * vmode->fbWidth + dxs) * VI_DISPLAY_PIX_SZ, GX_TRUE);
		}
	}
	
	GX_SetDrawDone();
	VIDEO_SetPostRetraceCallback(NULL);
}

static int WII_VideoInit(_THIS, SDL_PixelFormat *vformat)
{
	SetupGX();
	
	KEYBOARD_Init(NULL);
	MOUSE_Init();
	
	vformat->BitsPerPixel = 8;
	vformat->BytesPerPixel = 1;
	return(0);
}

static SDL_Rect
	mode_640x432 = {0,0,640,432},
	mode_640x480 = {0,0,640,480},
	mode_640x528 = {0,0,640,528};

static SDL_Rect *ntsc_modes[] = {
	&mode_640x480,
	&mode_640x432,
	NULL
};

static SDL_Rect *pal_modes[] = {
	&mode_640x528,
	NULL
};

static SDL_Rect **WII_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
{
	if (VIDEO_GetCurrentTvMode() == VI_PAL) {
		return(pal_modes);
	} else {
		return(ntsc_modes);
	}
}

static SDL_Surface *WII_SetVideoMode(_THIS, SDL_Surface *current, int width, int height, int bpp, Uint32 flags)
{
	SDL_Surface *video;
	
	if (width > vmode->fbWidth || height > vmode->xfbHeight) {
		SDL_SetError("%dx%d mode is too large", width, height);
		return(NULL);
	}
	
	if (bpp != 8) {
		SDL_SetError("%d bpp mode not supported", bpp);
		return(NULL);
	}
	
	free(this->hidden->buffer);
	SDL_FreeSurface(current);
	
	video = SDL_CreateRGBSurface(SDL_SWSURFACE, width, height, 8, 0, 0, 0, 0);
	
	if (video == NULL) {
		SDL_OutOfMemory();
		return(NULL);
	}
	
	video->flags |= SDL_HWPALETTE | SDL_FULLSCREEN;
	
	this->hidden->w = (video->w + 7) & ~7;
	this->hidden->h = (video->h + 3) & ~3;
	
	this->hidden->buffer = memalign(32, this->hidden->w * this->hidden->h + 32);
	
	DrawInit(this, video);
	WPAD_SetDataFormat(WPAD_CHAN_0, WPAD_FMT_BTNS_ACC_IR);
	WPAD_SetVRes(WPAD_CHAN_0, video->w + 96, video->h + 96);
	
	this->UpdateRects = WII_UpdateRects;
	return(video);
}

static void WII_UpdateRects(_THIS, int numrects, SDL_Rect *rects)
{
	VIDEO_SetPostRetraceCallback(NULL);
	GX_RedirectWriteGatherPipe(this->hidden->buffer);
	
	double *src1 = this->screen->pixels - sizeof(double);
	double *src2 = (void *)src1 + this->screen->pitch;
	double *src3 = (void *)src2 + this->screen->pitch;
	double *src4 = (void *)src3 + this->screen->pitch;
	
	int rowpitch = this->screen->pitch * 3;
	int rows = this->hidden->h >> 2;
	
	while (rows--) {
		int tiles = this->hidden->w >> 3;
		do {
			wgPipe->F64 = *++src1;
			wgPipe->F64 = *++src2;
			wgPipe->F64 = *++src3;
			wgPipe->F64 = *++src4;
		} while (--tiles);
		
		src1 = (void *)src1 + rowpitch;
		src2 = (void *)src2 + rowpitch;
		src3 = (void *)src3 + rowpitch;
		src4 = (void *)src4 + rowpitch;
	}
	
	GX_RestoreWriteGatherPipe();
	GX_InvalidateTexAll();
	
	VIDEO_SetPostRetraceCallback(VSyncCB);
	GX_WaitDrawDone();
}

static int WII_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors)
{
	int i;
	for (i = 0; i < ncolors; i++) {
		const SDL_Color *in = &colors[i];
		tlutmem[0][firstcolor + i] = (in->g << 8) | in->r;
		tlutmem[1][firstcolor + i] = ( 0xFF << 8) | in->b;
	}
	
	VIDEO_SetPostRetraceCallback(NULL);
	
	DCStoreRange(tlutmem[0], sizeof(tlutmem[0]));
	GX_LoadTlut(&tlutobj[0], GX_TLUT0);
	GX_LoadTexObj(&texobj[0], GX_TEXMAP0);
	
	DCStoreRange(tlutmem[1], sizeof(tlutmem[1]));
	GX_LoadTlut(&tlutobj[1], GX_TLUT1);
	GX_LoadTexObj(&texobj[1], GX_TEXMAP1);
	
	VIDEO_SetPostRetraceCallback(VSyncCB);
	return(1);
}

static void WII_VideoQuit(_THIS)
{
	KEYBOARD_Deinit();
	MOUSE_Deinit();
	
	GX_AbortFrame();
	
	VIDEO_SetBlack(true);
	VIDEO_Flush();
	VIDEO_WaitVSync();
}

static void WII_DeleteDevice(SDL_VideoDevice *device)
{
	SDL_free(device->hidden);
	SDL_free(device);
}

static SDL_VideoDevice *WII_CreateDevice(int devindex)
{
	SDL_VideoDevice *device;
	
	/* Initialize all variables that we clean on shutdown */
	device = (SDL_VideoDevice *)SDL_malloc(sizeof(SDL_VideoDevice));
	if ( device ) {
		SDL_memset(device, 0, (sizeof *device));
		device->hidden = (struct SDL_PrivateVideoData *)
			SDL_malloc((sizeof *device->hidden));
	}
	if ( (device == NULL) || (device->hidden == NULL) ) {
		SDL_OutOfMemory();
		if ( device ) {
			SDL_free(device);
		}
		return(0);
	}
	SDL_memset(device->hidden, 0, (sizeof *device->hidden));
	
	/* Set the function pointers */
	device->VideoInit = WII_VideoInit;
	device->ListModes = WII_ListModes;
	device->SetVideoMode = WII_SetVideoMode;
	device->CreateYUVOverlay = NULL;
	device->SetColors = WII_SetColors;
	device->UpdateRects = NULL;
	device->VideoQuit = WII_VideoQuit;
	device->AllocHWSurface = NULL;
	device->CheckHWBlit = NULL;
	device->FillHWRect = NULL;
	device->SetHWColorKey = NULL;
	device->SetHWAlpha = NULL;
	device->LockHWSurface = NULL;
	device->UnlockHWSurface = NULL;
	device->FlipHWSurface = NULL;
	device->FreeHWSurface = NULL;
	device->SetCaption = NULL;
	device->SetIcon = NULL;
	device->IconifyWindow = NULL;
	device->GrabInput = NULL;
	device->GetWMInfo = NULL;
	device->InitOSKeymap = WII_InitOSKeymap;
	device->PumpEvents = WII_PumpEvents;
	
	device->free = WII_DeleteDevice;
	return device;
}

static int WII_Available(void)
{
	return(1);
}

VideoBootStrap WII_bootstrap = {
	"wii", "Wii video driver",
	WII_Available, WII_CreateDevice
};

void WII_InitVideoSystem(void)
{
	VIDEO_Init();
	
	bool progressive = VIDEO_HaveComponentCable()
		&& (CONF_GetProgressiveScan() > 0);
	int format = CONF_GetVideo();
	
	if (progressive) {
		switch (format) {
			case CONF_VIDEO_PAL:
				if (CONF_GetEuRGB60() > 0)
					vmode = &TVEurgb60Hz480Prog;
				else vmode = &TVPal528Prog;
				break;
			case CONF_VIDEO_MPAL:
				vmode = &TVMpal480Prog;
				break;
			default:
				vmode = &TVNtsc480Prog;
		}
	} else {
		VIDEO_SetTrapFilter(true);
		switch (format) {
			case CONF_VIDEO_PAL:
				if (CONF_GetEuRGB60() > 0)
					vmode = &TVEurgb60Hz480IntDf;
				else vmode = &TVPal528IntDf;
				break;
			case CONF_VIDEO_MPAL:
				vmode = &TVMpal480IntDf;
				break;
			default:
				vmode = &TVNtsc480IntDf;
		}
	}
	
	VIDEO_Configure(vmode);
	
	xfb = SYS_AllocateFramebuffer(vmode);
	VIDEO_SetNextFramebuffer(xfb);
	
	VIDEO_SetBlack(false);
	VIDEO_Flush();
	VIDEO_WaitVSync();
}
