/* $Id$ */

/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

#include "../../stdafx.h"
#include "../../crashlog.h"
#include "../../string_func.h"
#include "../../gamelog.h"
#include "../../saveload/saveload.h"

#include <gccore.h>
#include <ogc/lwp_threads.h>
#include <ogc/libversion.h>

#define IS_ALIGNED(addr) (((intptr_t)(addr) % PPC_ALIGNMENT) == 0)
#define MAX_STACK_FRAMES 64

extern "C" {
	void ottd_exceptionhandler();
	void c_ottd_exceptionhandler(frame_context *);
	void __exception_sethandler(uint32, void (*)());
	void __exception_close(uint32);
	void __reload();
}

extern GXRModeObj *vmode;
static frame_context *exception_ctx = NULL;

static const char * const exception_name[NUM_EXCEPTIONS] = {
	"System Reset", "Machine Check", "DSI", "ISI",
	"Interrupt", "Alignment", "Program", "Floating Point",
	"Decrementer", "System Call", "Trace", "Performance",
	"IABR", "Reserved", "Thermal"
};

class CrashLogOGC : public CrashLog {
	/* virtual */ char *LogOSVersion(char *buffer, const char *last) const
	{
		return buffer + seprintf(buffer, last,
				"Operating system:\n"
				" Name:     libOGC\n"
				" Release:  %d.%d.%d\n",
				_V_MAJOR_, _V_MINOR_, _V_PATCH_
		);
	}

	/* virtual */ char *LogError(char *buffer, const char *last, const char *message) const
	{
		return buffer + seprintf(buffer, last,
				"Crash reason:\n"
				" Exception: %s (%u)\n"
				" Location:  %.8X\n"
				" Message:   %s\n\n",
				exception_name[exception_ctx->EXCPT_Number],
				exception_ctx->EXCPT_Number,
				exception_ctx->SRR0,
				message == NULL ? "<none>" : message
		);
	}

	/* virtual */ char *LogStacktrace(char *buffer, const char *last) const
	{
		buffer += seprintf(buffer, last, "Stack trace:\n");
		void **frame = (void **)exception_ctx->GPR[1];

		for (int i = 0; frame != NULL && i < MAX_STACK_FRAMES; i++) {
			void *ip = frame[1];
			if (ip == NULL) break;

			buffer += seprintf(buffer, last, " %.8X", (intptr_t)ip);
			if (i % 8 == 7) buffer += seprintf(buffer, last, "\n");

			void **next = (void **)frame[0];
			if (next <= frame || !IS_ALIGNED(next)) break;
			frame = next;
		}

		return buffer + seprintf(buffer, last, "\n\n");
	}

	/* virtual */ char *LogRegisters(char *buffer, const char *last) const
	{
		buffer += seprintf(buffer, last, "Registers:\n");

		for (int i = 0; i < 32; i++) {
			buffer += seprintf(buffer, last, " GPR%02d: %.8X", i, exception_ctx->GPR[i]);
			if (i % 4 == 3) buffer += seprintf(buffer, last, "\n");
		}

		buffer += seprintf(buffer, last,
			" PC: %.8X MSR: %.8X CR: %.8X LR: %.8X\n"
			" CTR: %.8X XER: %.8X\n",
			exception_ctx->SRR0,
			exception_ctx->SRR1,
			exception_ctx->CR,
			exception_ctx->LR,
			exception_ctx->CTR,
			exception_ctx->XER
		);

		return buffer + seprintf(buffer, last, "\n");
	}
};

static void WaitForReload()
{
	printf("Press RESET to continue.\n");
	while (!SYS_ResetButtonDown());

	VIDEO_SetBlack(true);
	VIDEO_Flush();
	VIDEO_WaitVSync();

	__reload();
}

static void HandleCrash()
{
	VIDEO_SetPostRetraceCallback(NULL);

	/* Avoid using the suggested static address */
	void *xfb = VIDEO_GetNextFramebuffer();
	CON_Init(xfb, 0, 0, vmode->fbWidth, vmode->xfbHeight, vmode->fbWidth * VI_DISPLAY_PIX_SZ);

	if (GamelogTestEmergency()) {
		printf("A serious fault condition occurred in the game. The game will shut down.\n");
		printf("As you loaded an emergency savegame no crash information will be generated.\n");
		WaitForReload();
	}

	if (SaveloadCrashWithMissingNewGRFs()) {
		printf("A serious fault condition occurred in the game. The game will shut down.\n");
		printf("As you loaded an savegame for which you do not have the required NewGRFs\n");
		printf("no crash information will be generated.\n");
		WaitForReload();
	}

	CrashLogOGC log;
	log.MakeCrashLog();
	WaitForReload();
}

static const uint32 _exception_to_handle[] = { EX_MACH_CHECK, EX_DSI, EX_ISI, EX_ALIGN, EX_PRG };

void c_ottd_exceptionhandler(frame_context *ctx)
{
	/* And they all hung happily ever after */
	for (const uint32 *i = _exception_to_handle; i != endof(_exception_to_handle); i++)
		__exception_close(*i);

	AUDIO_StopDMA();
	GX_AbortFrame();

	exception_ctx = ctx;
	__lwp_thread_stopmultitasking(HandleCrash);
}

/* static */ void CrashLog::InitialiseCrashLog()
{
	for (const uint32 *i = _exception_to_handle; i != endof(_exception_to_handle); i++)
		__exception_sethandler(*i, ottd_exceptionhandler);
}
